@@ -44,6 +44,7 @@ static const char * const btrfs_cmd_group_usage[] = {
 	"  -v|--verbose      increase verbosity of the subcommand\n"
 	"  -q|--quiet        print only errors\n"
 	"  --log <level>     set log level (default, info, verbose, debug, quiet)\n"
+	"  --dry-run         if supported, do not do any active/changing actions\n"
 	"\n"
 	"Options for the main command only:\n"
 	"  --help            print condensed help for all subcommands\n"
@@ -119,7 +120,7 @@ parse_command_token(const char *arg, const struct cmd_group *grp)
 	return cmd;
 }
 
-static void check_output_format(const struct cmd_struct *cmd)
+static void check_command_flags(const struct cmd_struct *cmd)
 {
 	if (cmd->next)
 		return;
@@ -129,6 +130,11 @@ static void check_output_format(const struct cmd_struct *cmd)
 			output_format_name(bconf.output_format));
 		exit(1);
 	}
+
+	if (bconf.dry_run && !(cmd->flags & CMD_DRY_RUN)) {
+		error("--dry-run option not supported for %s\n", cmd->token);
+		exit(1);
+	}
 }
 
 static void handle_help_options_next_level(const struct cmd_struct *cmd,
@@ -165,7 +171,7 @@ int handle_command_group(const struct cmd_struct *cmd, int argc,
 	subcmd = parse_command_token(argv[0], cmd->next);
 
 	handle_help_options_next_level(subcmd, argc, argv);
-	check_output_format(subcmd);
+	check_command_flags(subcmd);
 
 	fixup_argv0(argv, subcmd->token);
 	return cmd_execute(subcmd, argc, argv);
@@ -215,7 +221,7 @@ static const char * const cmd_version_usage[] = {
 
 static int cmd_version(const struct cmd_struct *unused, int argc, char **argv)
 {
-	printf("%s\n", PACKAGE_STRING);
+	help_builtin_features("");
 	return 0;
 }
 static DEFINE_SIMPLE_COMMAND(version, "version");
@@ -287,6 +293,8 @@ static int handle_global_options(int argc, char **argv)
 		{ "verbose", no_argument, NULL, 'v' },
 		{ "quiet", no_argument, NULL, 'q' },
 		{ "log", required_argument, NULL, OPT_LOG },
+		{ "param", required_argument, NULL, GETOPT_VAL_PARAM },
+		{ "dry-run", no_argument, NULL, GETOPT_VAL_DRY_RUN },
 		{ NULL, 0, NULL, 0}
 	};
 	int shift;
@@ -312,6 +320,12 @@ static int handle_global_options(int argc, char **argv)
 		case OPT_LOG:
 			handle_log_level(optarg);
 			break;
+		case GETOPT_VAL_PARAM:
+			bconf_save_param(optarg);
+			break;
+		case GETOPT_VAL_DRY_RUN:
+			bconf_set_dry_run();
+			break;
 		case 'v':
 			bconf_be_verbose();
 			break;
