@@ -21,6 +21,11 @@
 #ifndef __BTRFS_CHECK_MODE_ORIGINAL_H__
 #define __BTRFS_CHECK_MODE_ORIGINAL_H__
 
+#include "kerncompat.h"
+#include "kernel-lib/rbtree.h"
+#include "kernel-lib/list.h"
+#include "kernel-shared/uapi/btrfs_tree.h"
+#include "common/extent-cache.h"
 #include "common/rbtree-utils.h"
 
 struct extent_backref {
@@ -189,8 +194,7 @@ struct unaligned_extent_rec_t {
 #define I_ERR_INVALID_GEN		(1U << 20)
 #define I_ERR_INVALID_NLINK		(1U << 21)
 #define I_ERR_INVALID_XATTR		(1U << 22)
-/* Ram_bytes mismatch for non-compressed data extents. */
-#define I_ERR_RAM_BYTES_MISMATCH	(1U << 23)
+#define I_ERR_DEPRECATED_FREE_INO	(1U << 23)
 
 struct inode_record {
 	struct list_head backrefs;
@@ -218,7 +222,6 @@ struct inode_record {
 	u64 extent_end;
 	struct rb_root holes;
 	struct list_head mismatch_dir_hash;
-	struct list_head mismatch_ram_bytes;
 
 	u32 refs;
 };
@@ -235,11 +238,6 @@ struct mismatch_dir_hash_record {
 	/* namebuf follows here */
 };
 
-struct mismatch_ram_bytes_record {
-	struct list_head list;
-	struct btrfs_key key;
-};
-
 struct root_backref {
 	struct list_head list;
 	unsigned int found_dir_item:1;
