@@ -111,7 +111,6 @@
 #include "common/cpu-utils.h"
 #include "common/messages.h"
 #include "common/task-utils.h"
-#include "common/path-utils.h"
 #include "common/help.h"
 #include "common/parse-utils.h"
 #include "common/string-utils.h"
@@ -122,6 +121,7 @@
 #include "common/extent-tree-utils.h"
 #include "common/root-tree-utils.h"
 #include "common/clear-cache.h"
+#include "common/utils.h"
 #include "cmds/commands.h"
 #include "check/repair.h"
 #include "mkfs/common.h"
@@ -337,7 +337,7 @@ static int create_image_file_range(struct btrfs_trans_handle *trans,
 		error("remaining length not sectorsize aligned: %llu", len);
 		return -EINVAL;
 	}
-	ret = btrfs_record_file_extent(trans, root, ino, inode, bytenr,
+	ret = btrfs_convert_file_extent(trans, root, ino, inode, bytenr,
 				       disk_bytenr, len);
 	if (ret < 0)
 		return ret;
@@ -426,8 +426,8 @@ static int migrate_one_reserved_range(struct btrfs_trans_handle *trans,
 			break;
 
 		/* Now handle extent item and file extent things */
-		ret = btrfs_record_file_extent(trans, root, ino, inode, cur_off,
-					       key.objectid, key.offset);
+		ret = btrfs_convert_file_extent(trans, root, ino, inode, cur_off,
+					        key.objectid, key.offset);
 		if (ret < 0)
 			break;
 		/* Finally, insert csum items */
@@ -438,7 +438,7 @@ static int migrate_one_reserved_range(struct btrfs_trans_handle *trans,
 		/* Don't forget to insert hole */
 		hole_len = cur_off - hole_start;
 		if (hole_len) {
-			ret = btrfs_record_file_extent(trans, root, ino, inode,
+			ret = btrfs_convert_file_extent(trans, root, ino, inode,
 					hole_start, 0, hole_len);
 			if (ret < 0)
 				break;
@@ -455,7 +455,7 @@ static int migrate_one_reserved_range(struct btrfs_trans_handle *trans,
 	 *                   | Hole |
 	 */
 	if (range_end(range) - hole_start > 0)
-		ret = btrfs_record_file_extent(trans, root, ino, inode,
+		ret = btrfs_convert_file_extent(trans, root, ino, inode,
 				hole_start, 0, range_end(range) - hole_start);
 	return ret;
 }
@@ -1022,13 +1022,14 @@ static int init_btrfs(struct btrfs_mkfs_config *cfg, struct btrfs_root *root,
 			     BTRFS_FIRST_FREE_OBJECTID);
 
 	/* subvol for fs image file */
-	ret = btrfs_make_subvolume(trans, CONV_IMAGE_SUBVOL_OBJECTID);
+	ret = btrfs_make_subvolume(trans, CONV_IMAGE_SUBVOL_OBJECTID, false);
 	if (ret < 0) {
 		error("failed to create subvolume image root: %d", ret);
 		goto err;
 	}
 	/* subvol for data relocation tree */
-	ret = btrfs_make_subvolume(trans, BTRFS_DATA_RELOC_TREE_OBJECTID);
+	ret = btrfs_make_subvolume(trans, BTRFS_DATA_RELOC_TREE_OBJECTID,
+				   false);
 	if (ret < 0) {
 		error("failed to create DATA_RELOC root: %d", ret);
 		goto err;
@@ -1373,7 +1374,8 @@ static int do_convert(const char *devname, u32 convert_flags, u32 nodesize,
 	btrfs_sb_committed = true;
 
 	root = open_ctree_fd(fd, devname, 0,
-			     OPEN_CTREE_WRITES | OPEN_CTREE_TEMPORARY_SUPER);
+			     OPEN_CTREE_WRITES | OPEN_CTREE_TEMPORARY_SUPER |
+			     OPEN_CTREE_EXCLUSIVE);
 	if (!root) {
 		error("unable to open ctree for finalization");
 		goto fail;
@@ -1867,6 +1869,10 @@ static const char * const convert_usage[] = {
 	OPTLINE("-O|--features LIST", "comma separated list of filesystem features"),
 	OPTLINE("--no-progress", "show only overview, not the detailed progress"),
 	"",
+	"General:",
+	OPTLINE("--version", "print the btrfs-convert version, builtin features and exit"),
+	OPTLINE("--help", "print this help and exit"),
+	"",
 	"Supported filesystems:",
 	"\text2/3/4: "
 #if BTRFSCONVERT_EXT2
@@ -1908,12 +1914,12 @@ int BOX_MAIN(convert)(int argc, char *argv[])
 
 	cpu_detect_flags();
 	hash_init_accel();
+	btrfs_config_init();
 	btrfs_assert_feature_buf_size();
-	printf("btrfs-convert from %s\n\n", PACKAGE_STRING);
 
 	while(1) {
 		enum { GETOPT_VAL_NO_PROGRESS = GETOPT_VAL_FIRST, GETOPT_VAL_CHECKSUM,
-			GETOPT_VAL_UUID };
+			GETOPT_VAL_UUID, GETOPT_VAL_VERSION };
 		static const struct option long_options[] = {
 			{ "no-progress", no_argument, NULL,
 				GETOPT_VAL_NO_PROGRESS },
@@ -1931,7 +1937,8 @@ int BOX_MAIN(convert)(int argc, char *argv[])
 			{ "copy-label", no_argument, NULL, 'L' },
 			{ "uuid", required_argument, NULL, GETOPT_VAL_UUID },
 			{ "nodesize", required_argument, NULL, 'N' },
-			{ "help", no_argument, NULL, GETOPT_VAL_HELP},
+			{ "help", no_argument, NULL, GETOPT_VAL_HELP },
+			{ "version", no_argument, NULL, GETOPT_VAL_VERSION },
 			{ NULL, 0, NULL, 0 }
 		};
 		int c = getopt_long(argc, argv, "dinN:rl:LpO:", long_options, NULL);
@@ -2023,11 +2030,18 @@ int BOX_MAIN(convert)(int argc, char *argv[])
 					strncpy_null(fsid, optarg, sizeof(fsid));
 				}
 				break;
+			case GETOPT_VAL_VERSION:
+				help_builtin_features("btrfs-convert, part of ");
+				ret = 0;
+				goto success;
 			case GETOPT_VAL_HELP:
 			default:
 				usage(&convert_cmd, c != GETOPT_VAL_HELP);
 		}
 	}
+
+	printf("btrfs-convert from %s\n\n", PACKAGE_STRING);
+
 	set_argv0(argv);
 	if (check_argc_exact(argc - optind, 1)) {
 		usage(&convert_cmd, 1);
@@ -2071,5 +2085,6 @@ int BOX_MAIN(convert)(int argc, char *argv[])
 	}
 	if (ret)
 		return 1;
+success:
 	return 0;
 }
