@@ -41,7 +41,6 @@
 #include "kernel-shared/volumes.h"
 #include "kernel-shared/transaction.h"
 #include "kernel-shared/zoned.h"
-#include "kernel-shared/uuid-tree.h"
 #include "crypto/hash.h"
 #include "common/defs.h"
 #include "common/internal.h"
@@ -428,28 +427,50 @@ static const char * const mkfs_usage[] = {
 	OPTLINE("-d|--data PROFILE", "data profile, raid0, raid1, raid1c3, raid1c4, raid5, raid6, raid10, dup or single"),
 	OPTLINE("-m|--metadata PROFILE", "metadata profile, values like for data profile"),
 	OPTLINE("-M|--mixed","mix metadata and data together"),
+	"",
 	"Features:",
 	OPTLINE("--csum TYPE", ""),
 	OPTLINE("--checksum TYPE", "checksum algorithm to use, crc32c (default), xxhash, sha256, blake2"),
-	OPTLINE("-n|--nodesize SIZE", "size of btree nodes"),
-	OPTLINE("-s|--sectorsize SIZE", "data block size (may not be mountable by current kernel)"),
+	OPTLINE("-n|--nodesize SIZE", "size of btree nodes (default 16K)"),
+	OPTLINE("-s|--sectorsize SIZE", "data block size (default 4K, may not be mountable by current kernel if CPU page size is different)"),
 	OPTLINE("-O|--features LIST", "comma separated list of filesystem features (use '-O list-all' to list features)"),
-	OPTLINE("-L|--label LABEL", "set the filesystem label"),
+	OPTLINE("-L|--label LABEL", "set the filesystem label (maximum length 255)"),
 	OPTLINE("-U|--uuid UUID", "specify the filesystem UUID (must be unique for a filesystem with multiple devices)"),
-	OPTLINE("--device-uuid UUID", "Specify the filesystem device UUID (a.k.a sub-uuid) (for single device filesystem only)"),
+	OPTLINE("--device-uuid UUID", "specify the filesystem device UUID (a.k.a sub-uuid) (for single device filesystem only)"),
+	"",
 	"Creation:",
 	OPTLINE("-b|--byte-count SIZE", "set size of each device to SIZE (filesystem size is sum of all device sizes)"),
-	OPTLINE("-r|--rootdir DIR", "copy files from DIR to the image root directory"),
+	OPTLINE("-r|--rootdir DIR", "copy files from DIR to the image root directory, can be combined with --subvol"),
+	OPTLINE("--compress ALGO[:LEVEL]", "compress files by algorithm and level, ALGO can be 'no' (default), zstd, lzo, zlib"),
+	OPTLINE("", "Built-in:"),
+#if COMPRESSION_ZSTD
+	OPTLINE("", "- ZSTD: yes (levels 1..15)"),
+#else
+	OPTLINE("", "- ZSTD: no"),
+#endif
+#if COMPRESSION_LZO
+	OPTLINE("", "- LZO: yes"),
+#else
+	OPTLINE("", "- LZO: no"),
+#endif
+	OPTLINE("", "- ZLIB: yes (levels 1..9)"),
+	OPTLINE("-u|--subvol TYPE:SUBDIR", "create SUBDIR as subvolume rather than normal directory, can be specified multiple times"),
+	OPTLINE("", "TYPE is one of:"),
+	OPTLINE("", "- rw - (default) create a writeable subvolume in SUBDIR"),
+	OPTLINE("", "- ro - create the subvolume as read-only"),
+	OPTLINE("", "- default - the SUBDIR will be a subvolume and also set as default (can be specified only once)"),
+	OPTLINE("", "- defalut-ro - like 'default' and is created as read-only subvolume (can be specified only once)"),
 	OPTLINE("--shrink", "(with --rootdir) shrink the filled filesystem to minimal size"),
 	OPTLINE("-K|--nodiscard", "do not perform whole device TRIM"),
 	OPTLINE("-f|--force", "force overwrite of existing filesystem"),
+	"",
 	"General:",
 	OPTLINE("-q|--quiet", "no messages except errors"),
 	OPTLINE("-v|--verbose", "increase verbosity level, default is 1"),
-	OPTLINE("-V|--version", "print the mkfs.btrfs version and exit"),
+	OPTLINE("-V|--version", "print the mkfs.btrfs version, builtin features and exit"),
 	OPTLINE("--help", "print this help and exit"),
+	"",
 	"Deprecated:",
-	OPTLINE("-l|--leafsize SIZE", "removed in 6.0, use --nodesize"),
 	OPTLINE("-R|--runtime-features LIST", "removed in 6.3, use -O|--features"),
 	NULL
 };
@@ -881,7 +902,8 @@ static int setup_quota_root(struct btrfs_fs_info *fs_info)
 	}
 	ret = btrfs_create_root(trans, fs_info, BTRFS_QUOTA_TREE_OBJECTID);
 	if (ret < 0) {
-		error("failed to create quota root: %d (%m)", ret);
+		errno = -ret;
+		error("failed to create quota root: %m");
 		goto fail;
 	}
 	quota_root = fs_info->quota_root;
@@ -893,7 +915,8 @@ static int setup_quota_root(struct btrfs_fs_info *fs_info)
 	ret = btrfs_insert_empty_item(trans, quota_root, &path, &key,
 				      sizeof(*qsi));
 	if (ret < 0) {
-		error("failed to insert qgroup status item: %d (%m)", ret);
+		errno = -ret;
+		error("failed to insert qgroup status item: %m");
 		goto fail;
 	}
 
@@ -917,7 +940,8 @@ static int setup_quota_root(struct btrfs_fs_info *fs_info)
 	/* Currently mkfs will only create one subvolume */
 	ret = insert_qgroup_items(trans, fs_info, BTRFS_FS_TREE_OBJECTID);
 	if (ret < 0) {
-		error("failed to insert qgroup items: %d (%m)", ret);
+		errno = -ret;
+		error("failed to insert qgroup items: %m");
 		goto fail;
 	}
 
@@ -932,7 +956,8 @@ static int setup_quota_root(struct btrfs_fs_info *fs_info)
 	if (simple) {
 		ret = touch_root_subvol(fs_info);
 		if (ret) {
-			error("failed to touch root dir for simple quota accounting %d (%m)", ret);
+			errno = -ret;
+			error("failed to touch root dir for simple quota accounting: %m");
 			goto fail;
 		}
 	}
@@ -943,12 +968,15 @@ static int setup_quota_root(struct btrfs_fs_info *fs_info)
 	 */
 	ret = qgroup_verify_all(fs_info);
 	if (ret < 0) {
-		error("qgroup rescan failed: %d (%m)", ret);
+		errno = -ret;
+		error("qgroup rescan failed: %m");
 		return ret;
 	}
 	ret = repair_qgroups(fs_info, &qgroup_repaired, true);
-	if (ret < 0)
-		error("failed to fill qgroup info: %d (%m)", ret);
+	if (ret < 0) {
+		errno = -ret;
+		error("failed to fill qgroup info: %m");
+	}
 	return ret;
 fail:
 	btrfs_abort_transaction(trans, ret);
@@ -1014,6 +1042,125 @@ static void *prepare_one_device(void *ctx)
 	return NULL;
 }
 
+static int parse_compression(const char *str, enum btrfs_compression_type *type,
+			     unsigned int *level)
+{
+	const char *colon;
+	size_t type_size;
+	unsigned int default_level, max_level;
+
+	*level = 0;
+	if (strcmp(str, "no") == 0) {
+		*type = BTRFS_COMPRESS_NONE;
+		return 0;
+	}
+
+	colon = strstr(str, ":");
+
+	if (colon)
+		type_size = colon - str;
+	else
+		type_size = strlen(str);
+
+	if (strncmp(str, "zlib", type_size) == 0) {
+		*type = BTRFS_COMPRESS_ZLIB;
+		max_level = ZLIB_BTRFS_MAX_LEVEL;
+		default_level = ZLIB_BTRFS_DEFAULT_LEVEL;
+	} else if (strncmp(str, "lzo", type_size) == 0) {
+#if COMPRESSION_LZO
+		*type = BTRFS_COMPRESS_LZO;
+		max_level = 1;
+		default_level = 1;
+#else
+		error("lzo support not compiled in");
+		return 1;
+#endif
+	} else if (strncmp(str, "zstd", type_size) == 0) {
+#if COMPRESSION_ZSTD
+		*type = BTRFS_COMPRESS_ZSTD;
+		max_level = ZSTD_BTRFS_MAX_LEVEL;
+		default_level = ZSTD_BTRFS_DEFAULT_LEVEL;
+#else
+		error("zstd support not compiled in");
+		return 1;
+#endif
+	} else {
+		return 1;
+	}
+
+	if (colon) {
+		u64 tmplevel = arg_strtou64(colon + 1);
+
+		if (tmplevel > UINT_MAX)
+			return 1;
+
+		if (tmplevel == 0)
+			*level = default_level;
+		else if (tmplevel > max_level) {
+			error("compression level %llu out of range [1..%u]",
+			      tmplevel, max_level);
+			return 1;
+		} else {
+			*level = tmplevel;
+		}
+	}
+
+	return 0;
+}
+
+static int parse_subvolume(const char *path, struct list_head *subvols,
+			   bool *has_default_subvol)
+{
+	struct rootdir_subvol *subvol;
+	char *colon;
+	bool valid_prefix = false;
+
+	subvol = calloc(1, sizeof(struct rootdir_subvol));
+	if (!subvol) {
+		error_msg(ERROR_MSG_MEMORY, NULL);
+		return 1;
+	}
+
+	colon = strstr(optarg, ":");
+
+	if (colon) {
+		if (!string_has_prefix(optarg, "default:")) {
+			subvol->is_default = true;
+			valid_prefix = true;
+		} else if (!string_has_prefix(optarg, "ro:")) {
+			subvol->readonly = true;
+			valid_prefix = true;
+		} else if (!string_has_prefix(optarg, "rw:")) {
+			subvol->readonly = false;
+			valid_prefix = true;
+		} else if (!string_has_prefix(optarg, "default-ro:")) {
+			subvol->is_default = true;
+			subvol->readonly = true;
+			valid_prefix = true;
+		}
+	}
+
+	if (arg_copy_path(subvol->dir, valid_prefix ? colon + 1 : optarg,
+			  sizeof(subvol->dir))) {
+		error("--subvol path too long");
+		free(subvol);
+		return 1;
+	}
+
+	if (subvol->is_default) {
+		if (*has_default_subvol) {
+			error("default subvolume can only be specified once");
+			free(subvol);
+			return 1;
+		}
+		*has_default_subvol = true;
+	}
+
+	list_add_tail(&subvol->list, subvols);
+
+	return 0;
+}
+
 int BOX_MAIN(mkfs)(int argc, char **argv)
 {
 	char *file;
@@ -1055,6 +1202,11 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 	char *label = NULL;
 	int nr_global_roots = sysconf(_SC_NPROCESSORS_ONLN);
 	char *source_dir = NULL;
+	struct rootdir_subvol *rds;
+	bool has_default_subvol = false;
+	enum btrfs_compression_type compression = BTRFS_COMPRESS_NONE;
+	unsigned int compression_level = 0;
+	LIST_HEAD(subvols);
 
 	cpu_detect_flags();
 	hash_init_accel();
@@ -1068,6 +1220,7 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 			GETOPT_VAL_CHECKSUM,
 			GETOPT_VAL_GLOBAL_ROOTS,
 			GETOPT_VAL_DEVICE_UUID,
+			GETOPT_VAL_COMPRESS,
 		};
 		static const struct option long_options[] = {
 			{ "byte-count", required_argument, NULL, 'b' },
@@ -1076,7 +1229,6 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 			{ "checksum", required_argument, NULL,
 				GETOPT_VAL_CHECKSUM },
 			{ "force", no_argument, NULL, 'f' },
-			{ "leafsize", required_argument, NULL, 'l' },
 			{ "label", required_argument, NULL, 'L'},
 			{ "metadata", required_argument, NULL, 'm' },
 			{ "mixed", no_argument, NULL, 'M' },
@@ -1085,6 +1237,7 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 			{ "data", required_argument, NULL, 'd' },
 			{ "version", no_argument, NULL, 'V' },
 			{ "rootdir", required_argument, NULL, 'r' },
+			{ "subvol", required_argument, NULL, 'u' },
 			{ "nodiscard", no_argument, NULL, 'K' },
 			{ "features", required_argument, NULL, 'O' },
 			{ "runtime-features", required_argument, NULL, 'R' },
@@ -1094,6 +1247,8 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 			{ "quiet", 0, NULL, 'q' },
 			{ "verbose", 0, NULL, 'v' },
 			{ "shrink", no_argument, NULL, GETOPT_VAL_SHRINK },
+			{ "compress", required_argument, NULL,
+				GETOPT_VAL_COMPRESS },
 #if EXPERIMENTAL
 			{ "param", required_argument, NULL, GETOPT_VAL_PARAM },
 			{ "num-global-roots", required_argument, NULL, GETOPT_VAL_GLOBAL_ROOTS },
@@ -1102,7 +1257,7 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 			{ NULL, 0, NULL, 0}
 		};
 
-		c = getopt_long(argc, argv, "A:b:fl:n:s:m:d:L:R:O:r:U:VvMKq",
+		c = getopt_long(argc, argv, "A:b:fn:s:m:d:L:R:O:r:U:VvMKqu:",
 				long_options, NULL);
 		if (c < 0)
 			break;
@@ -1118,11 +1273,6 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 				}
 				data_profile_set = true;
 				break;
-			case 'l':
-				/* Deprecated in 4.0 */
-				error("--leafsize has been removed in 6.0, use --nodesize");
-				ret = 1;
-				goto error;
 			case 'n':
 				nodesize = arg_strtou64_with_suffix(optarg);
 				nodesize_forced = true;
@@ -1201,13 +1351,17 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 				bconf_be_verbose();
 				break;
 			case 'V':
-				printf("mkfs.btrfs, part of %s\n",
-						PACKAGE_STRING);
+				help_builtin_features("mkfs.btrfs, part of ");
 				goto success;
 			case 'r':
 				free(source_dir);
 				source_dir = strdup(optarg);
 				break;
+			case 'u':
+				ret = parse_subvolume(optarg, &subvols, &has_default_subvol);
+				if (ret)
+					exit(1);
+				break;
 			case 'U':
 				strncpy_null(fs_uuid, optarg, BTRFS_UUID_UNPARSED_SIZE);
 				break;
@@ -1217,6 +1371,12 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 			case 'q':
 				bconf_be_quiet();
 				break;
+			case GETOPT_VAL_COMPRESS:
+				if (parse_compression(optarg, &compression, &compression_level)) {
+					ret = 1;
+					goto error;
+				}
+				break;
 			case GETOPT_VAL_DEVICE_UUID:
 				strncpy_null(dev_uuid, optarg, BTRFS_UUID_UNPARSED_SIZE);
 				break;
@@ -1239,10 +1399,8 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 		}
 	}
 
-	if (bconf.verbose) {
-		printf("%s\n", PACKAGE_STRING);
-		printf("See %s for more information.\n\n", PACKAGE_URL);
-	}
+	pr_verbose(LOG_DEFAULT, "%s\n", PACKAGE_STRING);
+	pr_verbose(LOG_DEFAULT, "See %s for more information.\n\n", PACKAGE_URL);
 
 	if (!sectorsize)
 		sectorsize = (u32)SZ_4K;
@@ -1272,6 +1430,75 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 		ret = 1;
 		goto error;
 	}
+	if (!list_empty(&subvols) && source_dir == NULL) {
+		error("option --subvol must be used with --rootdir");
+		ret = 1;
+		goto error;
+	}
+
+	if (source_dir) {
+		char *canonical = realpath(source_dir, NULL);
+
+		if (!canonical) {
+			error("could not get canonical path to %s", source_dir);
+			ret = 1;
+			goto error;
+		}
+
+		free(source_dir);
+		source_dir = canonical;
+	} else {
+		if (compression != BTRFS_COMPRESS_NONE) {
+			error("--compression must be used with --rootdir");
+			ret = 1;
+			goto error;
+		}
+	}
+
+	list_for_each_entry(rds, &subvols, list) {
+		char path[PATH_MAX];
+		struct rootdir_subvol *rds2;
+
+		if (path_cat_out(path, source_dir, rds->dir)) {
+			error("path invalid: %s", path);
+			ret = 1;
+			goto error;
+		}
+
+		if (!realpath(path, rds->full_path)) {
+			error("could not get canonical path: %s", rds->dir);
+			ret = 1;
+			goto error;
+		}
+
+		if (!path_exists(rds->full_path)) {
+			error("subvolume path does not exist: %s", rds->dir);
+			ret = 1;
+			goto error;
+		}
+
+		if (!path_is_dir(rds->full_path)) {
+			error("subvolume is not a directory: %s", rds->dir);
+			ret = 1;
+			goto error;
+		}
+
+		if (!path_is_in_dir(source_dir, rds->full_path)) {
+			error("subvolume %s is not a child of %s", rds->dir, source_dir);
+			ret = 1;
+			goto error;
+		}
+
+		for (rds2 = list_first_entry(&subvols, struct rootdir_subvol, list);
+		     rds2 != rds;
+		     rds2 = list_next_entry(rds2, list)) {
+			if (strcmp(rds2->full_path, rds->full_path) == 0) {
+				error("subvolume specified more than once: %s", rds->dir);
+				ret = 1;
+				goto error;
+			}
+		}
+	}
 
 	if (*fs_uuid) {
 		uuid_t dummy_uuid;
@@ -1331,10 +1558,8 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 			exit(1);
 		}
 	} else if (zoned_model(file) == ZONED_HOST_MANAGED) {
-		if (bconf.verbose)
-			printf(
-	"Zoned: %s: host-managed device detected, setting zoned feature\n",
-			       file);
+		pr_verbose(LOG_DEFAULT, "zoned: %s: host-managed device detected, setting zoned feature\n",
+			   file);
 		opt_zoned = true;
 		features.incompat_flags |= BTRFS_FEATURE_INCOMPAT_ZONED;
 	}
@@ -1440,6 +1665,10 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 		goto error;
 	}
 
+	/* This is also fixed in kernel, but the flag has no real meaning anymore. */
+	if (nodesize > sysconf(_SC_PAGE_SIZE))
+		features.incompat_flags |= BTRFS_FEATURE_INCOMPAT_BIG_METADATA;
+
 	if (sectorsize < sizeof(struct btrfs_super_block)) {
 		error("sectorsize smaller than superblock: %u < %zu",
 				sectorsize, sizeof(struct btrfs_super_block));
@@ -1666,7 +1895,8 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 	}
 
 	oca.filename = file;
-	oca.flags = OPEN_CTREE_WRITES | OPEN_CTREE_TEMPORARY_SUPER;
+	oca.flags = OPEN_CTREE_WRITES | OPEN_CTREE_TEMPORARY_SUPER |
+		    OPEN_CTREE_EXCLUSIVE;
 	fs_info = open_ctree_fs_info(&oca);
 	if (!fs_info) {
 		error("open ctree failed");
@@ -1678,14 +1908,16 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 
 	ret = create_metadata_block_groups(root, mixed, &allocation);
 	if (ret) {
-		error("failed to create default block groups: %d", ret);
+		errno = -ret;
+		error("failed to create default block groups: %m");
 		goto error;
 	}
 
 	if (features.incompat_flags & BTRFS_FEATURE_INCOMPAT_RAID_STRIPE_TREE) {
 		ret = setup_raid_stripe_tree_root(fs_info);
 		if (ret < 0) {
-			error("failed to initialize raid-stripe-tree: %d (%m)", ret);
+			errno = -ret;
+			error("failed to initialize raid-stripe-tree: %m");
 			goto out;
 		}
 	}
@@ -1700,21 +1932,24 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 
 	ret = create_data_block_groups(trans, root, mixed, &allocation);
 	if (ret) {
-		error("failed to create default data block groups: %d", ret);
+		errno = -ret;
+		error("failed to create default data block groups: %m");
 		goto error;
 	}
 
 	if (features.incompat_flags & BTRFS_FEATURE_INCOMPAT_EXTENT_TREE_V2) {
 		ret = create_global_roots(trans, nr_global_roots);
 		if (ret) {
-			error("failed to create global roots: %d", ret);
+			errno = -ret;
+			error("failed to create global roots: %m");
 			goto error;
 		}
 	}
 
 	ret = make_root_dir(trans, root);
 	if (ret) {
-		error("failed to setup the root directory: %d", ret);
+		errno = -ret;
+		error("failed to setup the root directory: %m");
 		goto error;
 	}
 
@@ -1757,11 +1992,12 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 					prepare_ctx[i].file, dev_byte_count,
 					sectorsize, sectorsize, sectorsize);
 		if (ret) {
-			error("unable to add %s to filesystem: %d",
-			      prepare_ctx[i].file, ret);
+			errno = -ret;
+			error("unable to add %s to filesystem: %m",
+			      prepare_ctx[i].file);
 			goto error;
 		}
-		if (bconf.verbose >= 2) {
+		if (bconf.verbose >= LOG_INFO) {
 			struct btrfs_device *device;
 
 			device = container_of(fs_info->fs_devices->devices.next,
@@ -1777,7 +2013,8 @@ raid_groups:
 	ret = create_raid_groups(trans, root, data_profile,
 			 metadata_profile, mixed, &allocation);
 	if (ret) {
-		error("unable to create raid groups: %d", ret);
+		errno = -ret;
+		error("unable to create raid groups: %m");
 		goto out;
 	}
 
@@ -1809,9 +2046,11 @@ raid_groups:
 		goto out;
 	}
 
-	ret = btrfs_make_subvolume(trans, BTRFS_DATA_RELOC_TREE_OBJECTID);
+	ret = btrfs_make_subvolume(trans, BTRFS_DATA_RELOC_TREE_OBJECTID,
+				   false);
 	if (ret) {
-		error("unable to create data reloc tree: %d", ret);
+		errno = -ret;
+		error("unable to create data reloc tree: %m");
 		goto out;
 	}
 
@@ -1821,31 +2060,59 @@ raid_groups:
 		error_msg(ERROR_MSG_START_TRANS, "%m");
 		goto out;
 	}
-	ret = btrfs_rebuild_uuid_tree(fs_info);
-	if (ret < 0)
-		goto out;
-
-	ret = cleanup_temp_chunks(fs_info, &allocation, data_profile,
-				  metadata_profile, metadata_profile);
-	if (ret < 0) {
-		error("failed to cleanup temporary chunks: %d", ret);
-		goto out;
-	}
 
 	if (source_dir) {
 		pr_verbose(LOG_DEFAULT, "Rootdir from:       %s\n", source_dir);
-		ret = btrfs_mkfs_fill_dir(source_dir, root);
-		if (ret) {
-			error("error while filling filesystem: %d", ret);
+
+		trans = btrfs_start_transaction(root, 1);
+		if (IS_ERR(trans)) {
+			errno = -PTR_ERR(trans);
+			error_msg(ERROR_MSG_START_TRANS, "%m");
 			goto out;
 		}
+
+		pr_verbose(LOG_DEFAULT, "  Compress:         %s%s%s\n",
+			   compression == BTRFS_COMPRESS_ZSTD ? "zstd" :
+			   compression == BTRFS_COMPRESS_LZO ? "lzo" :
+			   compression == BTRFS_COMPRESS_ZLIB ? "zlib" : "no",
+			   compression_level > 0 ? ":" : "",
+			   compression_level > 0 ?
+				   pretty_size_mode(compression_level, UNITS_RAW) :
+				   "");
+
+		/* Print subvolumes now as btrfs_mkfs_fill_dir() deletes the list. */
+		list_for_each_entry(rds, &subvols, list) {
+			pr_verbose(LOG_DEFAULT, "  Subvolume (%s%s):  %s%s\n",
+				   rds->is_default ? "d" : "",
+				   rds->readonly ? "ro" : "rw",
+				   rds->is_default ? "" : " ",
+				   rds->dir);
+		}
+
+		ret = btrfs_mkfs_fill_dir(trans, source_dir, root,
+					  &subvols, compression,
+					  compression_level);
+		if (ret) {
+			errno = -ret;
+			error("error while filling filesystem: %m");
+			btrfs_abort_transaction(trans, ret);
+			goto out;
+		}
+
+		ret = btrfs_commit_transaction(trans, root);
+		if (ret) {
+			errno = -ret;
+			error_msg(ERROR_MSG_COMMIT_TRANS, "%m");
+			goto out;
+		}
+
 		if (shrink_rootdir) {
 			pr_verbose(LOG_DEFAULT, "  Shrink:           yes\n");
 			ret = btrfs_mkfs_shrink_fs(fs_info, &shrink_size,
 						   shrink_rootdir);
 			if (ret < 0) {
-				error("error while shrinking filesystem: %d",
-					ret);
+				errno = -ret;
+				error("error while shrinking filesystem: %m");
 				goto out;
 			}
 		} else {
@@ -1853,11 +2120,24 @@ raid_groups:
 		}
 	}
 
+	ret = btrfs_rebuild_uuid_tree(fs_info);
+	if (ret < 0)
+		goto out;
+
+	ret = cleanup_temp_chunks(fs_info, &allocation, data_profile,
+				  metadata_profile, metadata_profile);
+	if (ret < 0) {
+		errno = -ret;
+		error("failed to cleanup temporary chunks: %m");
+		goto out;
+	}
+
 	if (features.runtime_flags & BTRFS_FEATURE_RUNTIME_QUOTA ||
 	    features.incompat_flags & BTRFS_FEATURE_INCOMPAT_SIMPLE_QUOTA) {
 		ret = setup_quota_root(fs_info);
 		if (ret < 0) {
-			error("failed to initialize quota: %d (%m)", ret);
+			errno = -ret;
+			error("failed to initialize quota: %m");
 			goto out;
 		}
 	}
@@ -1930,8 +2210,8 @@ out:
 
 	if (!ret && close_ret) {
 		ret = close_ret;
-		error("failed to close ctree, the filesystem may be inconsistent: %d",
-		      ret);
+		errno = -ret;
+		error("failed to close ctree, filesystem may be inconsistent: %m");
 	}
 
 	btrfs_close_all_devices();
@@ -1946,6 +2226,14 @@ error:
 	free(label);
 	free(source_dir);
 
+	while (!list_empty(&subvols)) {
+		struct rootdir_subvol *head;
+
+		head = list_entry(subvols.next, struct rootdir_subvol, list);
+		list_del(&head->list);
+		free(head);
+	}
+
 	return !!ret;
 
 success:
